#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# Attempt to download and build PDFium from source. This may take very long.
# Only tested on Linux with glibc. Might not work on other platforms.

import os
import sys
import shutil
import argparse
import urllib.request
from os.path import join, abspath, dirname

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    SB_Dir,
    Libnames,
    DataTree,
    VerNamespace,
    PlatformNames,
    run_cmd,
    call_ctypesgen,
    set_version,
)


PatchDir       = join(SB_Dir, "patches")
DepotToolsDir  = join(SB_Dir, "depot_tools")
PDFiumDir      = join(SB_Dir, "pdfium")
PDFiumBuildDir = join(PDFiumDir, "out", "Default")
OutputDir      = join(DataTree, PlatformNames.sourcebuild)

DepotTools_URL = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
PDFium_URL     = "https://pdfium.googlesource.com/pdfium.git"

PdfiumMainPatches = [
    (join(PatchDir, "pdfium", "public_headers.patch"), PDFiumDir),
    (join(PatchDir, "pdfium", "shared_library.patch"), PDFiumDir),
]
PdfiumWinPatches = [
    (join(PatchDir, "pdfium", "win", "pdfium.patch"), PDFiumDir),
    (join(PatchDir, "pdfium", "win", "build.patch"), join(PDFiumDir, "build")),
]

DefaultConfig = {
    "is_debug": False,
    "treat_warnings_as_errors": False,
    # "clang_use_chrome_plugins": False,
    "pdf_is_standalone": True,
    "pdf_enable_v8": False,
    "pdf_enable_xfa": False,
    "pdf_use_skia": False,
}

if sys.platform.startswith("darwin"):
    DefaultConfig["mac_deployment_target"] = "10.11.0"
elif sys.platform.startswith("win32"):
    DefaultConfig["pdf_use_win32_gdi"] = True


def dl_depottools(do_update):
    
    if not os.path.isdir(SB_Dir):
        os.makedirs(SB_Dir)
    
    is_update = True
    
    if os.path.isdir(DepotToolsDir):
        if do_update:
            print("DepotTools: Revert and update ...")
            run_cmd([Git, "reset", "--hard", "HEAD"], cwd=DepotToolsDir)
            run_cmd([Git, "pull", DepotTools_URL], cwd=DepotToolsDir)
        else:
            print("DepotTools: Using existing repository as-is.")
            is_update = False
    else:
        print("DepotTools: Download ...")
        run_cmd([Git, "clone", "--depth", "1", DepotTools_URL, DepotToolsDir], cwd=SB_Dir)
    
    os.environ["PATH"] += os.pathsep + DepotToolsDir
    
    return is_update


def dl_pdfium(do_update, revision):
    
    is_sync = True
    
    if os.path.isdir(PDFiumDir):
        if do_update:
            print("PDFium: Revert / Sync  ...")
            run_cmd([GClient, "revert"], cwd=SB_Dir)
        else:
            is_sync = False
            print("PDFium: Using existing repository as-is.")
    else:
        print("PDFium: Download ...")
        run_cmd([GClient, "config", "--custom-var", "checkout_configuration=minimal", "--unmanaged", PDFium_URL], cwd=SB_Dir)
    
    if is_sync:
        run_cmd([GClient, "sync", "--revision", "origin/%s" % revision, "--no-history", "--shallow"], cwd=SB_Dir)
    
    return is_sync


def get_version_info():
    head_commit = run_cmd([Git, "rev-parse", "--short", "HEAD"], cwd=PDFiumDir, capture=True)
    refs_string = run_cmd([Git, "ls-remote", "--heads", PDFium_URL, "chromium/*"], cwd=None, capture=True)
    latest = refs_string.split("\n")[-1]
    tag_commit, ref = latest.split("\t")
    tag = ref.split("/")[-1]
    return head_commit, tag_commit[:7], tag


def update_version(head_commit, tag_commit, tag):
    
    is_sourcebuild  = VerNamespace["IS_SOURCEBUILD"]
    curr_libversion = VerNamespace["V_LIBPDFIUM"]
    
    print("Current head %s, latest tagged commit %s (%s)" % (head_commit, tag_commit, tag))
    if head_commit == tag_commit:
        new_libversion = tag
    else:
        new_libversion = head_commit
    
    if not is_sourcebuild:
        print("Switching from pre-built binaries to source build.")
        set_version("IS_SOURCEBUILD", True)
    if curr_libversion != new_libversion:
        set_version("V_LIBPDFIUM", new_libversion)


def _apply_patchset(patchset):
    for patch, cwd in patchset:
        run_cmd([Git, "apply", "-v", patch], cwd=cwd)

def patch_pdfium():
    _apply_patchset(PdfiumMainPatches)
    if sys.platform.startswith("win32"):
        _apply_patchset(PdfiumWinPatches)
        shutil.copy(join(PatchDir, "pdfium", "win", "resources.rc"), join(PDFiumDir, "resources.rc"))


def configure(config):
    if not os.path.exists(PDFiumBuildDir):
        os.makedirs(PDFiumBuildDir)
    with open(join(PDFiumBuildDir, "args.gn"), "w") as args_handle:
        args_handle.write(config)
    run_cmd([GN, "gen", PDFiumBuildDir], cwd=PDFiumDir)


def build(target):
    run_cmd([Ninja, "-C", PDFiumBuildDir, target], cwd=PDFiumDir)


def find_lib(srcname=None, directory=PDFiumBuildDir):
    
    if srcname is not None:
        path = join(PDFiumBuildDir, srcname)
        if os.path.isfile(path):
            return path
        else:
            print("Warning: The file of given srcname does not exist.", file=sys.stderr)
    
    libpath = None
    for lname in Libnames:
        path = join(directory, lname)
        if os.path.isfile(path):
            libpath = path
    
    if libpath is None:
        raise RuntimeError("Build artifact not found.")
    
    return libpath


def pack(src_libpath, destname=None):
    
    if os.path.isdir(OutputDir):
        shutil.rmtree(OutputDir)
    os.makedirs(OutputDir)
    
    if destname is None:
        destname = os.path.basename(src_libpath)
    
    destpath = join(OutputDir, destname)
    shutil.copy(src_libpath, destpath)
    
    include_dir = join(OutputDir, "include")
    shutil.copytree(join(PDFiumDir, "public"), include_dir)
    
    call_ctypesgen(OutputDir, include_dir)
    shutil.rmtree(include_dir)


def _get_tool(tool, win_append):
    exe = join(DepotToolsDir, tool)
    if sys.platform.startswith("win32"):
        exe += "." + win_append
    return exe


def _serialise_config(config_dict):
    
    config_str = ""
    sep = ""
    
    for key, value in config_dict.items():
        config_str += sep + "%s = " % key
        if isinstance(value, bool):
            config_str += str(value).lower()
        elif isinstance(value, str):
            config_str += '"%s"' % value
        else:
            raise TypeError("Not sure how to serialise type %s" % type(value))
        sep = "\n"
    
    return config_str


def main(
        b_srcname = None,
        b_destname = None,
        b_update = False,
        b_revision = None,
        b_target = None,
    ):
    
    global Git
    Git = shutil.which("git")
    
    if b_revision is None:
        b_revision = "main"
    if b_target is None:
        b_target = "pdfium"
    # on Linux, rename the binary to `pdfium` to ensure it also works with older versions of ctypesgen
    if b_destname is None and sys.platform.startswith("linux"):
        b_destname = "pdfium"
    
    if sys.platform.startswith("win32"):
        os.environ["DEPOT_TOOLS_WIN_TOOLCHAIN"] = "0"
    
    dl_depottools(b_update)
    
    global GClient, GN, Ninja
    GClient = _get_tool("gclient", "bat")
    GN      = _get_tool("gn", "bat")
    Ninja   = _get_tool("ninja", "exe")
    
    pdfium_dl_done = dl_pdfium(b_update, b_revision)
    if pdfium_dl_done:
        patch_pdfium()
    
    update_version( *get_version_info() )
    
    config_dict = DefaultConfig.copy()
    config_str = _serialise_config(config_dict)
    print("\nBuild configuration:\n%s\n" % config_str)
    
    configure(config_str)
    build(b_target)
    libpath = find_lib(b_srcname)
    pack(libpath, b_destname)


def parse_args(argv):
    
    parser = argparse.ArgumentParser(
        description = "A script to automate building PDFium from source and generating bindings with ctypesgen.",
    )
    
    parser.add_argument(
        "--srcname", "-s",
        help = "Name of the generated PDFium binary file. This script tries to automatically find the binary, which should usually work. If it does not, however, this option may be used to explicitly provide the file name to look for.",
    )
    parser.add_argument(
        "--destname", "-d",
        help = "Rename the binary to a different filename.",
    )
    parser.add_argument(
        "--update", "-u",
        action = "store_true",
        help = "Update existing PDFium/DepotTools repositories, removing local changes.",
    )
    parser.add_argument(
        "--revision", "-r",
        help = "PDFium revision to check out (defaults to main).",
    )
    parser.add_argument(
        "--target", "-t",
        help = "PDFium build target (defaults to `pdfium`). Use `pdfium_all` to also build tests."
    )
    
    return parser.parse_args(argv)


def main_cli(argv=sys.argv[1:]):
    args = parse_args(argv)
    return main(
        b_srcname = args.srcname,
        b_destname = args.destname,
        b_update = args.update,
        b_revision = args.revision,
        b_target = args.target,
    )
    

if __name__ == "__main__":
    main_cli()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import logging
from os.path import join, dirname, abspath, isdir, isfile

lib_logger = logging.getLogger("pypdfium2")
lib_logger.addHandler(logging.StreamHandler())

TestDir     = dirname(abspath(__file__))
SourceTree  = dirname(TestDir)
ResourceDir = join(TestDir, "resources")
OutputDir   = join(TestDir, "output")

sys.path.insert(0, join(SourceTree, "setupsrc"))


class TestFiles:
    render        = join(ResourceDir, "render.pdf")
    encrypted     = join(ResourceDir, "encrypted.pdf")
    multipage     = join(ResourceDir, "multipage.pdf")
    toc           = join(ResourceDir, "toc.pdf")
    toc_viewmodes = join(ResourceDir, "toc_viewmodes.pdf")
    toc_maxdepth  = join(ResourceDir, "toc_maxdepth.pdf")
    toc_circular  = join(ResourceDir, "toc_circular.pdf")
    box_fallback  = join(ResourceDir, "box_fallback.pdf")
    text          = join(ResourceDir, "text.pdf")
    empty         = join(ResourceDir, "empty.pdf")
    images        = join(ResourceDir, "images.pdf")


ExpRenderPixels = (
    ( (0,   0  ), (255, 255, 255) ),
    ( (150, 180), (129, 212, 26 ) ),
    ( (150, 390), (42,  96,  153) ),
    ( (150, 570), (128, 0,   128) ),
)


def iterate_testfiles(skip_encrypted=True):
    encrypted = (TestFiles.encrypted, )
    for attr_name in dir(TestFiles):
        if attr_name.startswith("_"):
            continue
        member = getattr(TestFiles, attr_name)
        if skip_encrypted and member in encrypted:
            continue
        yield member


def test_testpaths():
    for dirpath in (TestDir, SourceTree, ResourceDir, OutputDir):
        assert isdir(dirpath)
    for filepath in iterate_testfiles(False):
        assert isfile(filepath)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause


def pagetext_type(value):
    
    if not value:
        return
    
    page_indices = []
    splitted = value.split(",")
    
    for page_or_range in splitted:
        
        if "-" in page_or_range:
            
            start, end = page_or_range.split("-")
            start = int(start) - 1
            end = int(end) - 1
            
            if start < end:
                pages = [i for i in range(start, end+1)]
            else:
                pages = [i for i in range(start, end-1, -1)]
            
            page_indices.extend(pages)
        
        else:
            
            page_indices.append(int(page_or_range) - 1)
    
    return page_indices



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import pytest
import pypdfium2 as pdfium
from pypdfium2._helpers._utils import *


@pytest.mark.parametrize(
    ["degrees", "const"],
    [
        (0,   0),
        (90,  1),
        (180, 2),
        (270, 3),
    ]
)
def test_rotation_conversion(degrees, const):
    assert RotationToConst[degrees] == const
    assert RotationToDegrees[const] == degrees


@pytest.mark.parametrize(
    ["use_alpha", "greyscale", "cl_src", "cl_dest", "cl_const"],
    [
        (False, False, "BGR",  "RGB",  pdfium.FPDFBitmap_BGR),
        (True,  False, "BGRA", "RGBA", pdfium.FPDFBitmap_BGRA),
        (False, True,  "L",    "L",    pdfium.FPDFBitmap_Gray),
        (True,  True,  "BGRA", "RGBA", pdfium.FPDFBitmap_BGRA),
    ]
)
def test_get_colourformat(use_alpha, greyscale, cl_src, cl_dest, cl_const):
    assert (cl_src, cl_const) == get_colourformat(use_alpha, greyscale)
    assert ColourMapping[cl_src] == cl_dest


@pytest.mark.parametrize(
    ["values", "cl_value", "use_alpha"],
    [
        ((255, 255, 255),      0xFFFFFFFF, False),
        ((255, 255, 255, 255), 0xFFFFFFFF, False),
        ((0,   255, 255, 255), 0xFF00FFFF, False),
        ((255, 0,   255, 255), 0xFFFF00FF, False),
        ((255, 255, 0,   255), 0xFFFFFF00, False),
        ((255, 255, 255, 0  ), 0x00FFFFFF, True ),
    ]
)
def test_colour_tohex(values, cl_value, use_alpha):
    assert colour_tohex(*values) == (cl_value, use_alpha)


@pytest.mark.parametrize(
    ["prefix", "mapping"],
    [
        ("PDFDEST_VIEW_", ViewmodeMapping),
        ("FPDF_ERR_", ErrorMapping),
        ("FPDF_PAGEOBJ_", ObjtypeToName),
    ]
)
def test_const_tostring(prefix, mapping):
    viewmode_attrs = [a for a in dir(pdfium) if a.startswith(prefix)]
    assert len(viewmode_attrs) == len(mapping)
    for attr in viewmode_attrs:
        const = getattr(pdfium, attr)
        as_string = mapping[const]
        assert isinstance(const, int)
        assert isinstance(as_string, str)



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import io
import re
import shutil
import tempfile
import pytest
import PIL.Image
from os.path import join, abspath
import pypdfium2 as pdfium
from ..conftest import TestFiles, ExpRenderPixels


def _check_general(pdf, n_pages=1):
    assert isinstance(pdf, pdfium.PdfDocument)
    assert len(pdf) == n_pages
    version = pdf.get_version()
    assert isinstance(version, int)
    assert 10 < version < 30


def _check_render(pdf):
    
    page = pdf.get_page(0)
    pil_image = page.render_topil()
    page.close()
    
    assert pil_image.mode == "RGB"
    assert pil_image.size == (595, 842)
    for pos, exp_value in ExpRenderPixels:
        assert pil_image.getpixel(pos) == exp_value
    
    return pil_image


@pytest.fixture
def open_filepath_native():
    pdf = pdfium.PdfDocument(TestFiles.render)
    assert pdf._file_access is pdfium.FileAccess.NATIVE
    _check_general(pdf)
    yield _check_render(pdf)
    pdf.close()


@pytest.fixture
def open_bytes():
    
    with open(TestFiles.render, "rb") as buffer:
        bytedata = buffer.read()
    
    assert isinstance(bytedata, bytes)
    pdf = pdfium.PdfDocument(bytedata)
    
    _check_general(pdf)
    yield _check_render(pdf)
    
    pdf.close()


@pytest.fixture
def open_buffer():
    
    buffer = open(TestFiles.render, "rb")
    pdf = pdfium.PdfDocument(buffer)
    
    _check_general(pdf)
    yield _check_render(pdf)
    
    pdf.close()
    assert buffer.closed is False
    buffer.close()


def test_opener_inputtypes(open_filepath_native, open_bytes, open_buffer):
    first_image = open_filepath_native
    other_images = (
        open_bytes,
        open_buffer,
    )
    assert isinstance(first_image, PIL.Image.Image)
    assert all(first_image == other for other in other_images)


def test_open_buffer_autoclose():
    
    buffer = open(TestFiles.render, "rb")
    pdf = pdfium.PdfDocument(buffer, autoclose=True)
    _check_general(pdf)
    
    pdf.close()
    assert buffer.closed is True


def test_open_filepath_buffer():
    
    pdf = pdfium.PdfDocument(TestFiles.render, file_access=pdfium.FileAccess.BUFFER)
    
    assert pdf._orig_input == TestFiles.render
    assert isinstance(pdf._actual_input, io.BufferedReader)
    assert pdf._autoclose is True
    _check_general(pdf)
    
    pdf.close()
    assert pdf._actual_input.closed is True


def test_open_filepath_bytes():
    
    pdf = pdfium.PdfDocument(TestFiles.render, file_access=pdfium.FileAccess.BYTES)
    
    assert pdf._orig_input == TestFiles.render
    assert isinstance(pdf._actual_input, bytes)
    _check_general(pdf)
    
    pdf.close()


def test_open_encrypted():
    
    buffer = open(TestFiles.encrypted, "rb")
    bytedata = buffer.read()
    buffer.seek(0)
    
    test_cases = [
        (TestFiles.encrypted, "test_owner"),
        (TestFiles.encrypted, "test_user"),
        (bytedata, "test_owner"),
        (bytedata, "test_user"),
        (buffer, "test_owner"),
        (buffer, "test_user"),
    ]
    
    for input_data, password in test_cases:
        pdf = pdfium.PdfDocument(input_data, password)
        _check_general(pdf)
        pdf.close()
        if input_data is buffer:
            buffer.seek(0)
    
    buffer.close()
    
    with pytest.raises(pdfium.PdfiumError, match=re.escape("Loading the document failed (PDFium: Incorrect password error)")):
        pdf = pdfium.PdfDocument(TestFiles.encrypted, "wrong_password")
        pdf.close()


@pytest.mark.parametrize(
    "file_access",
    [pdfium.FileAccess.NATIVE, pdfium.FileAccess.BYTES, pdfium.FileAccess.BUFFER]
)
def test_open_nonencrypted_with_password(file_access):
    pdf = pdfium.PdfDocument(TestFiles.render, password="irrelevant", file_access=file_access)
    _check_general(pdf)
    pdf.close()


def test_open_nonascii():
    
    tempdir = tempfile.TemporaryDirectory(prefix="pypdfium2_")
    nonascii_file = join(tempdir.name, "tên file chứakýtự éèáàçß 发短信.pdf")
    shutil.copy(TestFiles.render, nonascii_file)
    
    pdf = pdfium.PdfDocument(nonascii_file)
    _check_general(pdf)
    pdf.close()
    
    tempdir.cleanup()


def test_open_new():
    
    dest_pdf = pdfium.PdfDocument.new()
    
    assert isinstance(dest_pdf, pdfium.PdfDocument)
    assert isinstance(dest_pdf.raw, pdfium.FPDF_DOCUMENT)
    assert dest_pdf.raw is dest_pdf._orig_input is dest_pdf._actual_input
    assert dest_pdf._rendering_input is None
    assert dest_pdf._ld_data is None
    
    assert dest_pdf.get_version() is None
    
    src_pdf = pdfium.PdfDocument(TestFiles.multipage)
    pdfium.FPDF_ImportPages(dest_pdf.raw, src_pdf.raw, b"1, 3", 0)
    assert len(dest_pdf) == 2
    
    for g in (dest_pdf, src_pdf): g.close()


def test_open_invalid():
    with pytest.raises(TypeError, match=re.escape("__init__() missing 1 required positional argument: 'input_data'")):
        pdf = pdfium.PdfDocument()
    with pytest.raises(TypeError, match=re.escape("Invalid input type 'int'")):
        pdf = pdfium.PdfDocument(123)
    with pytest.raises(FileNotFoundError, match=re.escape("File does not exist: '%s'" % abspath("invalid/path"))):
        pdf = pdfium.PdfDocument("invalid/path")
    with pytest.raises(FileNotFoundError, match=re.escape("File does not exist: '%s'" % abspath("invalid/path"))):
        pdf = pdfium.PdfDocument("invalid/path", file_access=pdfium.FileAccess.BUFFER)
    with pytest.raises(ValueError, match=re.escape("An invalid file access strategy was given: 'abcd'")):
        pdf = pdfium.PdfDocument(TestFiles.empty, file_access="abcd")


def test_object_hierarchy():
    
    pdf = pdfium.PdfDocument(TestFiles.empty)
    assert isinstance(pdf, pdfium.PdfDocument)
    assert isinstance(pdf.raw, pdfium.FPDF_DOCUMENT)
    
    page = pdf.get_page(0)
    assert isinstance(page, pdfium.PdfPage)
    assert isinstance(page.raw, pdfium.FPDF_PAGE)
    assert page.pdf is pdf
    
    textpage = page.get_textpage()
    assert isinstance(textpage, pdfium.PdfTextPage)
    assert isinstance(textpage.raw, pdfium.FPDF_TEXTPAGE)
    assert textpage.page is page
    
    searcher = textpage.search("abcd")
    assert isinstance(searcher, pdfium.PdfTextSearcher)
    assert isinstance(searcher.raw, pdfium.FPDF_SCHHANDLE)
    assert searcher.textpage is textpage
    
    for g in (searcher, textpage, page, pdf): g.close()


def test_doc_extras():
    
    with pdfium.PdfDocument(TestFiles.empty, file_access=pdfium.FileAccess.BUFFER) as pdf:
        assert isinstance(pdf, pdfium.PdfDocument)
        assert len(pdf) == 1
        page = pdf[0]
        assert isinstance(page, pdfium.PdfPage)
        page.close()
    assert isinstance(pdf._actual_input, io.BufferedReader)
    assert pdf._actual_input.closed is True
    
    with pdfium.PdfDocument.new() as pdf:
        
        assert isinstance(pdf, pdfium.PdfDocument)
        assert len(pdf) == 0
        
        sizes = [(50, 100), (100, 150), (150, 200)]
        for size in sizes:
            page = pdf.new_page(*size)
            page.close()
        for size, page in zip(sizes, pdf):
            assert isinstance(page, pdfium.PdfPage)
            assert size == page.get_size()
            page.close()
        
        del pdf[0]
        page = pdf[0]
        assert page.get_size() == (100, 150)
        page.close()



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import time
import copy
import shutil
import argparse
from os.path import (
    join,
    exists,
    abspath,
    dirname,
)

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    run_cmd,
    set_version,
    get_version_ns,
    SourceTree,
    RepositoryURL,
    Changelog,
    ChangelogStaging,
    VersionFile,
    VerNamespace,
)
from pl_setup.update_pdfium import get_latest_version

AutoreleaseDir = join(SourceTree, "autorelease")
MajorUpdate = join(AutoreleaseDir, "majorupdate.txt")
BetaUpdate = join(AutoreleaseDir, "betaupdate.txt")


def run_local(*args, **kws):
    return run_cmd(*args, **kws, cwd=SourceTree)


def do_versioning(latest):
    
    # sourcebuild version changes must never be checked into version control
    # autorelease can't work with that state because it needs information about the previous release for its version changes
    assert not VerNamespace["IS_SOURCEBUILD"]
    assert VerNamespace["V_LIBPDFIUM"].isnumeric()
    
    v_libpdfium = int(VerNamespace["V_LIBPDFIUM"])
    
    # current libpdfium version mustn't be larger than determined latest, or something went badly wrong
    assert v_libpdfium <= latest
    
    if exists(MajorUpdate):
        set_version("V_MAJOR", VerNamespace["V_MAJOR"]+1)
        set_version("V_MINOR", 0)
        set_version("V_PATCH", 0)
        set_version("V_LIBPDFIUM", str(latest))
        os.remove(MajorUpdate)
    elif v_libpdfium < latest:
        set_version("V_MINOR", VerNamespace["V_MINOR"]+1)
        set_version("V_PATCH", 0)
        set_version("V_LIBPDFIUM", str(latest))
    else:
        # TODO check if there are any new commits compared to the previous release - if not, stop
        set_version("V_PATCH", VerNamespace["V_PATCH"]+1)
    
    if exists(BetaUpdate):
        v_beta = VerNamespace["V_BETA"]
        if v_beta is None:
            v_beta = 0
        set_version("V_BETA", v_beta+1)
        os.remove(BetaUpdate)


def get_summary():
    
    with open(ChangelogStaging, "r") as fh:
        content = fh.read()
        pos = content.index("\n", content.index("# Changelog")) + 1
        header = content[:pos].strip() + "\n"
        devel_msg = content[pos:].strip()
        if devel_msg:
            devel_msg += "\n"

    with open(ChangelogStaging, "w") as fh:
        fh.write(header)
    
    return devel_msg


def log_changes(summary, prev_ns, curr_ns):
    
    pdfium_msg = "## %s (%s)\n\n- " % (curr_ns["V_PYPDFIUM2"], time.strftime("%Y-%m-%d"))
    if prev_ns["V_LIBPDFIUM"] != curr_ns["V_LIBPDFIUM"]:
        pdfium_msg += "Updated PDFium from `%s` to `%s`" % (prev_ns["V_LIBPDFIUM"], curr_ns["V_LIBPDFIUM"])
    else:
        pdfium_msg += "No PDFium update"
    pdfium_msg += " (autorelease)."
    
    with open(Changelog, "r") as fh:
        content = fh.read()
        pos = content.index("\n", content.index("# Changelog")) + 1
        part_a = content[:pos].strip() + "\n"
        part_b = content[pos:].strip()
        content = part_a + "\n\n" + pdfium_msg + "\n" + summary + "\n\n" + part_b + "\n"
    with open(Changelog, "w") as fh:
        fh.write(content)


def register_changes(curr_ns, publish=False):
    
    # TODO move push commands into the release workflow - run them only after successful wheel building, to avoid possible breakage
    
    run_local([Git, "add", AutoreleaseDir, VersionFile, Changelog, ChangelogStaging])
    run_local([Git, "commit", "-m", "[autorelease] update changelog and version file"])
    if publish:
        run_local([Git, "push"])
    
    run_local([Git, "tag", "-a", curr_ns["V_PYPDFIUM2"], "-m", "Autorelease"])
    if publish:
        run_local([Git, "push", "--tags"])
    
    run_local([Git, "checkout", "stable"])
    run_local([Git, "reset", "--hard", "main"])
    if publish:
        run_local([Git, "push"])
    
    run_local([Git, "checkout", "main"])


def link_for_tag(tag):
    return RepositoryURL + "/tree/%s" % tag


def make_releasenotes(summary, prev_ns, curr_ns):
    
    prev_ver = prev_ns["V_PYPDFIUM2"]
    curr_ver = curr_ns["V_PYPDFIUM2"]
    
    relnotes = "Release %s\n\n" % curr_ver
    relnotes += "## Changes\n\n"
    relnotes += "### Manual Summary\n\n"
    if summary:
        relnotes += summary + "\n"
    relnotes += "### Git History\n\nCommits between "
    relnotes += "[`%s`](%s) and [`%s`](%s) " % (prev_ver, link_for_tag(prev_ver), curr_ver, link_for_tag(curr_ver))
    relnotes += "(latest commit first):\n\n"
    relnotes += run_local([Git, "log", "%s..%s" % (prev_ver, curr_ver), "--pretty=format:* %H %s"], capture=True)
    relnotes += "\n"
    
    with open(join(SourceTree, "RELEASE.md"), "w") as fh:
        fh.write(relnotes)


def main():
    
    # The `--checkin` and `--publish` options exist to avoid accidential repository changes or even pushes, and to simplify testing this script.
    parser = argparse.ArgumentParser(
        description = "Automatic update script for pypdfium2, to be run in the CI release workflow."
    )
    parser.add_argument(
        "--checkin",
        action = "store_true",
        help = "Allow running modifying git commands (commit, tag, reset)."
    )
    parser.add_argument(
        "--publish",
        action = "store_true",
        help = "Allow pushing changes to the public repository. Takes no effect if `--checkin` is not given.",
    )
    args = parser.parse_args()
    
    global Git
    Git = shutil.which("git")
    
    prev_ns = copy.deepcopy(VerNamespace)
    latest = get_latest_version()
    do_versioning(latest)
    curr_ns = get_version_ns()
    
    summary = get_summary()
    log_changes(summary, prev_ns, curr_ns)
    if args.checkin:
        register_changes(curr_ns, publish=args.publish)
    make_releasenotes(summary, prev_ns, curr_ns)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import ctypes
import pypdfium2._pypdfium as pdfium
from pypdfium2._helpers.misc import raise_error


class BufferDataHolder:
    
    def __init__(self, reader_func, buffer):
        self.reader_func = reader_func
        self.buffer = buffer
    
    def close(self):
        id(self.reader_func)
        id(self.buffer)


class ByteDataHolder:
    
    def __init__(self, bytedata):
        self.bytedata = bytedata
    
    def close(self):
        id(self.bytedata)


class ReaderClass:
    
    def __init__(self, buffer):
        self._buffer = buffer
    
    def __call__(self, _, position, p_buf, size):
        c_buf = (ctypes.c_char * size).from_address( ctypes.addressof(p_buf.contents) )
        self._buffer.seek(position)
        self._buffer.readinto(c_buf)
        return 1


def is_input_buffer(maybe_buffer):
    return all( callable(getattr(maybe_buffer, a, None)) for a in ("seek", "tell", "read", "readinto") )


def open_pdf_buffer(buffer, password=None):
    
    buffer.seek(0, 2)
    file_len = buffer.tell()
    buffer.seek(0)
    
    FuncType = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(None), ctypes.c_ulong, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_ulong)
    
    fileaccess = pdfium.FPDF_FILEACCESS()
    fileaccess.m_FileLen = file_len
    fileaccess.m_GetBlock = FuncType( ReaderClass(buffer) )
    
    pdf = pdfium.FPDF_LoadCustomDocument(ctypes.byref(fileaccess), password)
    ld_data = BufferDataHolder(fileaccess.m_GetBlock, buffer)
    
    return pdf, ld_data


def open_pdf_bytes(bytedata, password=None):
    pdf = pdfium.FPDF_LoadMemDocument64(bytedata, len(bytedata), password)
    ld_data = ByteDataHolder(bytedata)
    return pdf, ld_data


def open_pdf(input_data, password=None):
    
    ld_data = None
    if isinstance(input_data, str):
        pdf = pdfium.FPDF_LoadDocument(input_data, password)
    elif isinstance(input_data, bytes):
        pdf, ld_data = open_pdf_bytes(input_data, password)
    elif is_input_buffer(input_data):
        pdf, ld_data = open_pdf_buffer(input_data, password)
    else:
        raise TypeError("Invalid input type '%s'" % type(input_data).__name__)
    
    if pdfium.FPDF_GetPageCount(pdf) < 1:
        raise_error("Loading the document failed")
    
    return pdf, ld_data



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import platform
import sysconfig
import setuptools
from os.path import (
    join,
    abspath,
    dirname,
    basename,
)

sys.path.insert(0, join(dirname(abspath(__file__)), "setupsrc"))
from pl_setup.packaging_base import (
    PlatformNames,
    SourceTree,
    SetupTargetVar,
)


def packaging_handler():
    
    target = os.environ.get(SetupTargetVar, None)
    if target in (None, "auto"):
        return True
    
    from pl_setup.setup_base import mkwheel, SetupKws
    
    if target == "sdist":
        setuptools.setup(**SetupKws)
    elif hasattr(PlatformNames, target):
        mkwheel( getattr(PlatformNames, target) )
    else:
        raise ValueError("Invalid deployment target '%s'" % target)
    
    return False


def install_handler():
    
    from pl_setup import check_deps
    
    def check_presetup():
        StatusFile = join(SourceTree, "data", ".presetup_done.txt")
        if os.path.exists(StatusFile):
            return False
        else:
            with open(StatusFile, "w") as fh:
                fh.write("")
            return True
    
    W_Presetup = check_presetup()
    if W_Presetup: check_deps.main()
    
    from pl_setup import update_pdfium  #, build_pdfium
    from pl_setup.setup_base import mkwheel
    
    class HostPlatform:
        
        def __init__(self):
            
            plat_name = sysconfig.get_platform().lower()
            for char in ("-", "."):
                plat_name = plat_name.replace(char, "_")
            self.plat_name = plat_name
            
            self.libc_name = None
            if self.plat_name.startswith("linux"):
                self.libc_name = platform.libc_ver()[0]
        
        def is_platform(self, start, end):
            if self.plat_name.startswith(start):
                if self.plat_name.endswith(end):
                    return True
            return False
        
        def is_libc(self, *libc_names):
            return any(self.libc_name == l for l in libc_names)
    
    def _setup(pl_name):
        if W_Presetup: update_pdfium.main( [basename(pl_name)] )
        mkwheel(pl_name)
    
    host = HostPlatform()
    
    if host.is_platform("macosx", "arm64"):
        _setup(PlatformNames.darwin_arm64)
    elif host.is_platform("macosx", "x86_64"):
        _setup(PlatformNames.darwin_x64)
    elif host.is_platform("linux", "armv7l"):
        _setup(PlatformNames.linux_arm32)
    elif host.is_platform("linux", "aarch64"):
        _setup(PlatformNames.linux_arm64)
    elif host.is_platform("linux", "x86_64") and host.is_libc("glibc"):
        _setup(PlatformNames.linux_x64)
    elif host.is_platform("linux", "i686") and host.is_libc("glibc"):
        _setup(PlatformNames.linux_x86)
    elif host.is_platform("linux", "x86_64") and host.is_libc("musl", ""):
        _setup(PlatformNames.musllinux_x64)
    elif host.is_platform("linux", "i686") and host.is_libc("musl", ""):
        _setup(PlatformNames.musllinux_x86)
    elif host.is_platform("win", "arm64"):
        _setup(PlatformNames.windows_arm64)
    elif host.is_platform("win", "amd64"):
        _setup(PlatformNames.windows_x64)
    elif host.is_platform("win32", ""):
        _setup(PlatformNames.windows_x86)
    else:
        raise RuntimeError("No pre-built binaries available for platform '%s' with libc implementation '%s'. You can attempt a source build, but it's unlikely to work out due to binary toolchain requirements of PDFium's build system. Doing cross-compilation or using a different build system might be possible, though. Please get in touch with the project maintainers." % (host.plat_name, host.libc_name))
        # if W_Presetup: build_pdfium.main()
        # mkwheel(PlatformNames.sourcebuild)


def main():
    cont = packaging_handler()
    if cont: install_handler()

if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import argparse
from os.path import dirname, abspath

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    run_cmd,
    clean_artefacts,
    PlatformNames,
    AllPlatNames,
    SourceTree,
    SetupTargetVar,
)


def _run_build(args):
    run_cmd([sys.executable, "-m", "build", "--skip-dependency-check", "--no-isolation"] + args, cwd=SourceTree, env=os.environ)


def main():
    
    # This script cleans up any artefacts in src/ and currently does not restore them
    # i. e. if you were using an editable install, you'll have to re-run it (or emplace the artefacts manually)
    
    parser = argparse.ArgumentParser(
        description = "Craft sdist and wheels for pypdfium2, using `python3 -m build`. (This script does not take any arguments.)",
    )
    args = parser.parse_args()
    
    pl_names = AllPlatNames.copy()
    pl_names.remove(PlatformNames.sourcebuild)
    
    clean_artefacts()
    os.environ[SetupTargetVar] = "sdist"
    _run_build(["--sdist"])
    clean_artefacts()
    
    for plat in pl_names:
        os.environ[SetupTargetVar] = plat
        _run_build(["--wheel"])
        clean_artefacts()


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause 

import copy
import shutil
import tempfile
from os.path import join
import pypdfium2.version as pdfium_ver
from pl_setup import setup_base
import pl_setup.packaging_base as pkg_base
# import pl_setup.update_pdfium as fpdf_up
import pl_setup.build_pdfium as fpdf_build

try:
    from importlib.metadata import version as dist_version
except ImportError:
    from importlib_metadata import version as dist_version


def _check_namespace(V_MAJOR, V_MINOR, V_PATCH, V_BETA, V_PYPDFIUM2, V_LIBPDFIUM, IS_SOURCEBUILD):
    
    for m in (V_MAJOR, V_MINOR, V_PATCH):
        assert isinstance(m, int)
    for m in (V_PYPDFIUM2, V_LIBPDFIUM):
        assert isinstance(m, str)
    
    assert isinstance(V_BETA, (type(None), int))
    assert isinstance(IS_SOURCEBUILD, bool)
    
    assert V_PYPDFIUM2.count(".") == 2
    if not IS_SOURCEBUILD:
        assert V_LIBPDFIUM.isnumeric()
    if V_BETA is not None:
        assert "b" in V_PYPDFIUM2
        assert V_BETA >= 1
    
    assert V_MAJOR >= 2 and V_MINOR >= 0 and V_PATCH >= 0
    if V_LIBPDFIUM.isnumeric():
        assert int(V_LIBPDFIUM) > 5000


def test_current_version():
    _check_namespace(**pkg_base.VerNamespace)
    assert setup_base.SetupKws["version"] == pkg_base.VerNamespace["V_PYPDFIUM2"]

def test_installed_version():
    assert dist_version("pypdfium2") == pdfium_ver.V_PYPDFIUM2


def _test_change(method, exp_items):
    
    exp_ns = copy.deepcopy(pkg_base.VerNamespace)
    for key, value in exp_items:
        exp_ns[key] = value
    
    method()
    
    assert exp_ns == pkg_base.VerNamespace
    new_ns = pkg_base.get_version_ns()
    exp_ns["V_PYPDFIUM2"] = new_ns["V_PYPDFIUM2"]
    assert exp_ns == new_ns


def test_setversion():
    
    # TODO test autorelease.do_versioning() and update_pdfium.handle_versions()
    
    tempdir = tempfile.TemporaryDirectory()
    orig_vfile = pkg_base.VersionFile
    tmp_vfile = join(tempdir.name, "tmp_versionfile.py")
    shutil.copy(orig_vfile, tmp_vfile)
    pkg_base.VersionFile = tmp_vfile
    
    exp_items = (
        ("IS_SOURCEBUILD", False),
        ("V_MAJOR", 2),
        ("V_MINOR", 0),
        ("V_PATCH", 0),
        ("V_BETA", None),
        ("V_LIBPDFIUM", "5005"),
    )
    for key, value in exp_items:
        pkg_base.set_version(key, value)
    
    _test_change(lambda: None, exp_items)
    _test_change(
        lambda: fpdf_build.update_version("abcdefg", "1234567", "5012"),
        [
            ("IS_SOURCEBUILD", True),
            ("V_LIBPDFIUM", "abcdefg"),
        ]
    )
    _test_change(
        lambda: fpdf_build.update_version("1234567", "1234567", "5012"),
        [
            ("V_LIBPDFIUM", "5012"),
        ]
    )
    
    pkg_base.VersionFile = orig_vfile
    pkg_base.VerNamespace = pkg_base.get_version_ns()
    tempdir.cleanup()



# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import os
import sys
import shutil
import sysconfig
import setuptools
from glob import glob
from os.path import join, abspath, dirname, basename
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import (
    Libnames,
    DataTree,
    ModuleDir,
    VerNamespace,
    PlatformNames,
    clean_artefacts,
)


def _copy_bindings(pl_name):
    
    files = [f for f in glob(join(DataTree, pl_name, '*')) if os.path.isfile(f)]
    assert len(files) == 2
    
    for src_path in files:
        dest_path = join(ModuleDir, basename(src_path))
        shutil.copy(src_path, dest_path)


def _get_linux_tag(arch):
    return "manylinux_2_17_%s.manylinux2014_%s" % (arch, arch)

def _get_musllinux_tag(arch):
    return "musllinux_1_2_%s" % (arch)


def _get_mac_tag(arch, *versions):
    
    assert len(versions) > 0
    
    template = "macosx_%s_%s"
    
    tag = ""
    sep = ""
    for v in versions:
        tag += sep + template % (v, arch)
        sep = "."
    
    return tag


def _get_tag(pl_name):
    # It looks like pip >= 20.3 now accepts macOS wheels with a lower major version than the host system (that is, at least for 10 on 11).
    # Anyway, even if it's fixed now, we definitely still need to support older releases of pip
    # We're intrigued why other major libraries (e. g. pikepdf, pymupdf - both using cibuildwheel) are not doing multi-version tagging for macOS while they still have compatibility tags for manylinux2014.
    if pl_name == PlatformNames.darwin_x64:
        return _get_mac_tag("x86_64", "10_11", "11_0", "12_0")
    elif pl_name == PlatformNames.darwin_arm64:
        return _get_mac_tag("arm64", "11_0", "12_0")
    elif pl_name == PlatformNames.linux_x64:
        return _get_linux_tag("x86_64")
    elif pl_name == PlatformNames.linux_x86:
        return _get_linux_tag("i686")
    elif pl_name == PlatformNames.linux_arm64:
        return _get_linux_tag("aarch64")
    elif pl_name == PlatformNames.linux_arm32:
        return _get_linux_tag("armv7l")
    elif pl_name == PlatformNames.musllinux_x64:
        return _get_musllinux_tag("x86_64")
    elif pl_name == PlatformNames.musllinux_x86:
        return _get_musllinux_tag("i686")
    elif pl_name == PlatformNames.windows_x64:
        return "win_amd64"
    elif pl_name == PlatformNames.windows_arm64:
        return "win_arm64"
    elif pl_name == PlatformNames.windows_x86:
        return "win32"
    elif pl_name == PlatformNames.sourcebuild:
        tag = sysconfig.get_platform()
        for char in ("-", "."):
            tag = tag.replace(char, "_")
        return tag
    else:
        raise ValueError("Unknown platform directory %s" % pl_name)


def _get_bdist(pl_name):
    
    class bdist (_bdist_wheel):
        
        def finalize_options(self, *args, **kws):
            _bdist_wheel.finalize_options(self, *args, **kws)
            self.plat_name_supplied = True
            self.root_is_pure = False
        
        def get_tag(self, *args, **kws):
            return "py3", "none", _get_tag(pl_name)
    
    return bdist


SetupKws = dict(
    version = VerNamespace["V_PYPDFIUM2"],
)


class BinaryDistribution (setuptools.Distribution):
    def has_ext_modules(self):
        return True


def mkwheel(pl_name):
    
    clean_artefacts()
    _copy_bindings(pl_name)
    
    setuptools.setup(
        package_data = {"": Libnames},
        cmdclass = {"bdist_wheel": _get_bdist(pl_name)},
        distclass = BinaryDistribution,
        **SetupKws,
    )



#! /usr/bin/env python3
# SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

import sys
import shutil
import importlib.util
from os.path import abspath, dirname

sys.path.insert(0, dirname(dirname(abspath(__file__))))
from pl_setup.packaging_base import run_cmd


PyPackages = (
    "build",
    "wheel",
    "ctypesgen",
    "setuptools",
    "setuptools-scm",
)
SysCommands = (
    "git",
    "gcc",
)


def _pip_install(pkg):
    run_cmd([sys.executable, "-m", "pip", "install", pkg], cwd=None)


def install_pydeps():
    
    for pkg in PyPackages:
        if not importlib.util.find_spec( pkg.replace("-", "_") ):
            _pip_install(pkg)
    
    # Uninstalling ctypesgen sometimes leaves parts behind. In this case, `find_spec()` would still consider the library installed, although the command is unavailable.
    if not shutil.which("ctypesgen"):
        _pip_install("ctypesgen")


def check_sysdeps():
    
    missing = []
    found = []
    
    for dep_name in SysCommands:
        if shutil.which(dep_name):
            found.append(dep_name)
        else:
            missing.append(dep_name)
    
    print("Found system dependencies: %s" % found)
    if len(missing) > 0:
        print("Missing system dependencies: %s" % missing)


def main():
    check_sysdeps()
    install_pydeps()


if __name__ == "__main__":
    main()
